import struct
import socket
import select
import sys
import logging

if sys.version_info[0] < 3:
  import serialize
else:
  from . import serialize
#   from rtde import serialize

DEFAULT_TIMEOUT = 1.0

LOGNAME = 'rtde'
_log = logging.getLogger(LOGNAME)


class Command:
    RTDE_REQUEST_PROTOCOL_VERSION = 86        # ascii V
    RTDE_GET_URCONTROL_VERSION = 118          # ascii v
    RTDE_TEXT_MESSAGE = 77                    # ascii M
    RTDE_DATA_PACKAGE = 85                    # ascii U
    RTDE_CONTROL_PACKAGE_SETUP_OUTPUTS = 79   # ascii O
    RTDE_CONTROL_PACKAGE_SETUP_INPUTS = 73    # ascii I
    RTDE_CONTROL_PACKAGE_START = 83           # ascii S
    RTDE_CONTROL_PACKAGE_PAUSE = 80           # ascii P

RTDE_PROTOCOL_VERSION_1 = 1
RTDE_PROTOCOL_VERSION_2 = 2

class ConnectionState:
    DISCONNECTED = 0
    CONNECTED = 1
    STARTED = 2
    PAUSED = 3

class RTDEException(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self):
        return repr(self.msg)

class RTDETimeoutException(RTDEException):
    def __init__(self, msg):
        super(RTDETimeoutException, self).__init__(msg)

class RTDE(object):
    def __init__(self, hostname, port=30004):
        self.hostname = hostname
        self.port = port
        self.__conn_state = ConnectionState.DISCONNECTED
        self.__sock = None
        self.__output_config = None
        self.__input_config = {}
        self.__skipped_package_count = 0
        self.__protocolVersion = RTDE_PROTOCOL_VERSION_1

    def connect(self):
        if self.__sock:
            return

        self.__buf = b'' # buffer data in binary format
        try:
            self.__sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.__sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.__sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            self.__sock.settimeout(DEFAULT_TIMEOUT)
            self.__skipped_package_count = 0
            self.__sock.connect((self.hostname, self.port))
            self.__conn_state = ConnectionState.CONNECTED
        except (socket.timeout, socket.error):
            self.__sock = None
            raise
        if not self.negotiate_protocol_version():
            raise RTDEException('Unable to negotiate protocol version')

    def disconnect(self):
        if self.__sock:
            self.__sock.close()
            self.__sock = None
        self.__conn_state = ConnectionState.DISCONNECTED

    def is_connected(self):
        return self.__conn_state is not ConnectionState.DISCONNECTED

    def get_controller_version(self):
        cmd = Command.RTDE_GET_URCONTROL_VERSION
        version = self.__sendAndReceive(cmd)
        if version:
            _log.info('Controller version: ' + str(version.major) + '.' + str(version.minor) + '.' + str(version.bugfix)+ '.' + str(version.build))
            if version.major == 3 and version.minor <= 2 and version.bugfix < 19171:
                _log.error("Please upgrade your controller to minimally version 3.2.19171")
                sys.exit()
            return version.major, version.minor, version.bugfix, version.build
        return None, None, None, None

    def negotiate_protocol_version(self):
        cmd = Command.RTDE_REQUEST_PROTOCOL_VERSION
        payload = struct.pack('>H', RTDE_PROTOCOL_VERSION_2)
        success = self.__sendAndReceive(cmd, payload)
        if success:
            self.__protocolVersion = RTDE_PROTOCOL_VERSION_2
        return success

    def send_input_setup(self, variables, types=[]):
        cmd = Command.RTDE_CONTROL_PACKAGE_SETUP_INPUTS
        payload = bytearray(','.join(variables), 'utf-8')
        result = self.__sendAndReceive(cmd, payload)
        if len(types)!=0 and not self.__list_equals(result.types, types):
            _log.error('Data type inconsistency for input setup: ' +
                     str(types) + ' - ' +
                     str(result.types))
            return None
        result.names = variables
        self.__input_config[result.id] = result
        return serialize.DataObject.create_empty(variables, result.id)

    def send_output_setup(self, variables, types=[], frequency=125):
        cmd = Command.RTDE_CONTROL_PACKAGE_SETUP_OUTPUTS
        payload = struct.pack('>d', frequency)
        payload = payload + (','.join(variables).encode('utf-8'))
        result = self.__sendAndReceive(cmd, payload)
        if len(types)!=0 and not self.__list_equals(result.types, types):
            _log.error('Data type inconsistency for output setup: ' +
                     str(types) + ' - ' +
                     str(result.types))
            return False
        result.names = variables
        self.__output_config = result
        return True

    def send_start(self):
        cmd = Command.RTDE_CONTROL_PACKAGE_START
        success = self.__sendAndReceive(cmd)
        if success:
            _log.info('RTDE synchronization started')
            self.__conn_state = ConnectionState.STARTED
        else:
            _log.error('RTDE synchronization failed to start')
        return success

    def send_pause(self):
        cmd = Command.RTDE_CONTROL_PACKAGE_PAUSE
        success = self.__sendAndReceive(cmd)
        if success:
            _log.info('RTDE synchronization paused')
            self.__conn_state = ConnectionState.PAUSED
        else:
            _log.error('RTDE synchronization failed to pause')
        return success

    def send(self, input_data):
        if self.__conn_state != ConnectionState.STARTED:
            _log.error('Cannot send when RTDE synchronization is inactive')
            return
        if not input_data.recipe_id in self.__input_config:
            _log.error('Input configuration id not found: ' + str(input_data.recipe_id))
            return
        config = self.__input_config[input_data.recipe_id]
        return self.__sendall(Command.RTDE_DATA_PACKAGE, config.pack(input_data))


    def receive(self, binary=False):
        """Recieve the latest data package.
        If muliple packages has been received, older ones are discarded
        and only the newest one will be returned. Will block untill a package
        is received or the connection is lost
        """
        if self.__output_config is None:
            raise RTDEException('Output configuration not initialized')
        if self.__conn_state != ConnectionState.STARTED:
            raise RTDEException('Cannot receive when RTDE synchronization is inactive')
        return self.__recv(Command.RTDE_DATA_PACKAGE, binary)

    def receive_buffered(self, binary = False, buffer_limit = None):
        """Recieve the next data package.
        If muliple packages has been received they are buffered and will
        be returned on subsequent calls to this function.
        Returns None if no data is available.
        """

        if self._RTDE__output_config is None:
            logging.error('Output configuration not initialized')
            return None

        try:
            while self.is_connected() and \
                  (buffer_limit == None or len(self.__buf) < buffer_limit) and \
                  self.__recv_to_buffer(0):
                pass
        except RTDEException as e:
            data = self.__recv_from_buffer(Command.RTDE_DATA_PACKAGE, binary)
            if data == None:
                raise e
        else:
            data = self.__recv_from_buffer(Command.RTDE_DATA_PACKAGE, binary)

        return data

    def send_message(self, message, source = "Python Client", type = serialize.Message.INFO_MESSAGE):
        cmd = Command.RTDE_TEXT_MESSAGE
        fmt = '>B%dsB%dsB' % (len(message), len(source))
        payload = struct.pack(fmt, len(message), message, len(source), source, type)
        return self.__sendall(cmd, payload)

    def __on_packet(self, cmd, payload):
        if cmd == Command.RTDE_REQUEST_PROTOCOL_VERSION:
            return self.__unpack_protocol_version_package(payload)
        elif cmd == Command.RTDE_GET_URCONTROL_VERSION:
            return self.__unpack_urcontrol_version_package(payload)
        elif cmd == Command.RTDE_TEXT_MESSAGE:
            return self.__unpack_text_message(payload)
        elif cmd == Command.RTDE_CONTROL_PACKAGE_SETUP_OUTPUTS:
            return self.__unpack_setup_outputs_package(payload)
        elif cmd == Command.RTDE_CONTROL_PACKAGE_SETUP_INPUTS:
            return self.__unpack_setup_inputs_package(payload)
        elif cmd == Command.RTDE_CONTROL_PACKAGE_START:
            return self.__unpack_start_package(payload)
        elif cmd == Command.RTDE_CONTROL_PACKAGE_PAUSE:
            return self.__unpack_pause_package(payload)
        elif cmd == Command.RTDE_DATA_PACKAGE:
            return self.__unpack_data_package(payload, self.__output_config)
        else:
            _log.error('Unknown package command: ' + str(cmd))

    def __sendAndReceive(self, cmd, payload=b''):
        if self.__sendall(cmd, payload):
            return self.__recv(cmd)
        else:
            return None

    def __sendall(self, command, payload=b''):
        fmt = '>HB'
        size = struct.calcsize(fmt) + len(payload)
        buf = struct.pack(fmt, size, command) + payload

        if self.__sock is None:
            _log.error('Unable to send: not connected to Robot')
            return False

        _, writable, _ = select.select([], [self.__sock], [], DEFAULT_TIMEOUT)
        if len(writable):
            self.__sock.sendall(buf)
            return True
        else:
            self.__trigger_disconnected()
            return False

    def has_data(self):
        timeout = 0
        readable, _, _ = select.select([self.__sock], [], [], timeout)
        return len(readable)!=0

    def __recv(self, command, binary=False):
        while self.is_connected():
            try:
                self.__recv_to_buffer(DEFAULT_TIMEOUT)
            except RTDETimeoutException:
                return None

            # unpack_from requires a buffer of at least 3 bytes
            while len(self.__buf) >= 3:
                # Attempts to extract a packet
                packet_header = serialize.ControlHeader.unpack(self.__buf)

                if len(self.__buf) >= packet_header.size:
                    packet, self.__buf = self.__buf[3:packet_header.size], self.__buf[packet_header.size:]
                    data = self.__on_packet(packet_header.command, packet)
                    if len(self.__buf) >= 3 and command == Command.RTDE_DATA_PACKAGE:
                        next_packet_header = serialize.ControlHeader.unpack(self.__buf)
                        if next_packet_header.command == command:
                            _log.debug('skipping package(1)')
                            self.__skipped_package_count += 1
                            continue
                    if packet_header.command == command:
                        if(binary):
                            return packet[1:]

                        return data
                    else:
                        _log.debug('skipping package(2)')
                else:
                    break
        raise RTDEException(' _recv() Connection lost ')

    def __recv_to_buffer(self, timeout):
        readable, _, xlist = select.select([self.__sock], [], [self.__sock], timeout)
        if len(readable):
            more = self.__sock.recv(4096)
            #When the controller stops while the script is running
            if len(more) == 0:
                _log.error('received 0 bytes from Controller, probable cause: Controller has stopped')
                self.__trigger_disconnected()  
                raise RTDEException('received 0 bytes from Controller')

            self.__buf = self.__buf + more
            return True

        if (len(xlist) or len(readable) == 0) and timeout != 0: # Effectively a timeout of timeout seconds
            _log.warning('no data received in last %d seconds ',timeout)
            raise RTDETimeoutException("no data received within timeout")

        return False


    def __recv_from_buffer(self, command, binary=False):
        # unpack_from requires a buffer of at least 3 bytes
        while len(self.__buf) >= 3:
            # Attempts to extract a packet
            packet_header = serialize.ControlHeader.unpack(self.__buf)

            if len(self.__buf) >= packet_header.size:
                packet, self.__buf = self.__buf[3:packet_header.size], self.__buf[packet_header.size:]
                data = self.__on_packet(packet_header.command, packet)
                if packet_header.command == command:
                    if(binary):
                        return packet[1:]

                    return data
                else:
                    _log.debug('skipping package(2)')
            else:
                return None

    def __trigger_disconnected(self):
        _log.info("RTDE disconnected")
        self.disconnect() #clean-up

    def __unpack_protocol_version_package(self, payload):
        if len(payload) != 1:
            _log.error('RTDE_REQUEST_PROTOCOL_VERSION: Wrong payload size')
            return None
        result = serialize.ReturnValue.unpack(payload)
        return result.success

    def __unpack_urcontrol_version_package(self, payload):
        if len(payload) != 16:
            _log.error('RTDE_GET_URCONTROL_VERSION: Wrong payload size')
            return None
        version = serialize.ControlVersion.unpack(payload)
        return version

    def __unpack_text_message(self, payload):
        if len(payload) < 1:
            _log.error('RTDE_TEXT_MESSAGE: No payload')
            return None
        if(self.__protocolVersion == RTDE_PROTOCOL_VERSION_1):
            msg = serialize.MessageV1.unpack(payload)
        else:
            msg = serialize.Message.unpack(payload)

        if(msg.level == serialize.Message.EXCEPTION_MESSAGE or
           msg.level == serialize.Message.ERROR_MESSAGE):
            _log.error(msg.source + ': ' + msg.message)
        elif msg.level == serialize.Message.WARNING_MESSAGE:
            _log.warning(msg.source + ': ' + msg.message)
        elif msg.level == serialize.Message.INFO_MESSAGE:
            _log.info(msg.source + ': ' + msg.message)

    def __unpack_setup_outputs_package(self, payload):
        if len(payload) < 1:
            _log.error('RTDE_CONTROL_PACKAGE_SETUP_OUTPUTS: No payload')
            return None
        output_config = serialize.DataConfig.unpack_recipe(payload)
        return output_config

    def __unpack_setup_inputs_package(self, payload):
        if len(payload) < 1:
            _log.error('RTDE_CONTROL_PACKAGE_SETUP_INPUTS: No payload')
            return None
        input_config = serialize.DataConfig.unpack_recipe(payload)
        return input_config

    def __unpack_start_package(self, payload):
        if len(payload) != 1:
            _log.error('RTDE_CONTROL_PACKAGE_START: Wrong payload size')
            return None
        result = serialize.ReturnValue.unpack(payload)
        return result.success

    def __unpack_pause_package(self, payload):
        if len(payload) != 1:
            _log.error('RTDE_CONTROL_PACKAGE_PAUSE: Wrong payload size')
            return None
        result = serialize.ReturnValue.unpack(payload)
        return result.success

    def __unpack_data_package(self, payload, output_config):
        if output_config is None:
            _log.error('RTDE_DATA_PACKAGE: Missing output configuration')
            return None
        output = output_config.unpack(payload)
        return output

    def __list_equals(self, l1, l2):
        if len(l1) != len(l2):
            return False
        for i in range(len((l1))):
            if l1[i] != l2[i]:
                return False
        return True
    
    @property
    def skipped_package_count(self):
        """The skipped package count, resets on connect"""
        return self.__skipped_package_count



import xml.etree.ElementTree as ET


class Recipe(object):
    __slots__=['key', 'names', 'types']
    @staticmethod
    def parse(recipe_node):
        rmd = Recipe()
        rmd.key = recipe_node.get('key')
        rmd.names = [f.get('name') for f in recipe_node.findall('field')]
        rmd.types = [f.get('type') for f in recipe_node.findall('field')]
        return rmd


class ConfigFile(object):
    def __init__(self, filename):
        self.__filename = filename
        tree = ET.parse(self.__filename)
        root = tree.getroot()
        recipes = [Recipe.parse(r) for r in root.findall('recipe')]
        self.__dictionary = dict()
        for r in recipes:
            self.__dictionary[r.key] = r
        
    def get_recipe(self, key):
        r = self.__dictionary[key]
        return r.names, r.types



from __future__ import annotations

from enum import Enum, auto
from typing import Union, List

import os
import time
import socket
import argparse

import logging

from .rtde import rtde
from .rtde import rtde_config


PORT_SEND = 30002
PORT_RECEIVE = 30004
CONFIG = "configuration.xml"


class MovementType(Enum):
    '''Defined how the robot should move.'''
    LINEAR = auto()
    QUICKEST = auto()


class UniversalRobot:
    '''A helper class to communicate with a Universal Robot.'''
    def __init__(self, host_ip: str) -> None:
        self._host: str = host_ip
        self._accel: float = 1.5
        self._vel: float = 1.5

        parser = argparse.ArgumentParser()
        parser.add_argument('--host', default=self._host,help='name of host to connect to (localhost)')
        parser.add_argument('--port', type=int, default=PORT_RECEIVE, help='port number (30004)')
        parser.add_argument('--samples', type=int, default=0,help='number of samples to record')
        parser.add_argument('--frequency', type=int, default=125, help='the sampling frequency in Herz')
        parser.add_argument('--config', default=os.path.join(os.path.dirname(__file__), CONFIG), help='data configuration file to use (record_configuration.xml)')
        parser.add_argument("--verbose", help="increase output verbosity", action="store_true")
        parser.add_argument("--buffered", help="Use buffered receive which doesn't skip data", action="store_true")
        parser.add_argument("--binary", help="save the data in binary format", action="store_true")
        self._args = parser.parse_args()

        if self._args.verbose:
            logging.basicConfig(level=logging.INFO)

    def set_accel(self, accel: float) -> None:
        '''Sets the acceleration for the robot.'''
        self._accel = accel
    
    def set_vel(self, vel: float) -> None:
        '''Sets the velcoity for the robot.'''
        self._vel = vel

    def set_freedrive(self, state=True) -> None:
        '''Sets the robot in freedrive mode until another request is sent.'''
        function_str = None

        if state:
            function_str = "def prog():\nfreedrive_mode()\nsleep(99999999)\nend\nprog()"
        else:
            function_str = "end_freedrive_mode()\n"

        self._send_to_robot(function_str)

    def _send_to_robot(self, function_str: str) -> None:
        '''Sends a function string to the robot.'''
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((self._host, PORT_SEND))
        s.send(function_str.encode())
        s.close()
    
    def move_to(self, target: Union[Pose, JointPosition], movement_type: MovementType = MovementType.QUICKEST, wait: bool = True) -> None:
        '''Moves the robot to the desiered pose or joint position, waits before continuing if the 'wait' flag is set.'''
        if isinstance(target, Pose) and movement_type is MovementType.LINEAR:
            self._send_to_robot(target.movel(self._accel, self._vel))
        else :
            self._send_to_robot(target.movej(self._accel, self._vel))

        # Wait for robot to reach the position
        if wait:
            move_done = False
            while not move_done:
                current = self.get_pose() if isinstance(target, Pose) else self.get_joint_position()
                move_done = target == current
                if not move_done:
                    time.sleep(0.5)

    def move_path(self, path_points: List[PathPoint], wait: bool = True) -> None:
        '''Given an array of PathPoints the robot moves smoothly between them with little delay. Waits before continuing.'''
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((self._host, PORT_SEND))

        function_str: str = "def path():\n"

        # Create the function string
        for point in path_points:
            if isinstance(point.target, JointPosition):
                function_str = function_str + point.target.movej(self._accel, self._vel)
            else:
                if point.movement_type == MovementType.QUICKEST:
                    function_str = function_str + point.target.movej(self._accel, self._vel)
                elif point.movement_type == MovementType.LINEAR:
                    function_str = function_str + point.target.movel(self._accel, self._vel)
                else:
                    print("[urpy][ERROR]: Unknown movement type.")

        function_str += "end\n"     # End the script
        function_str += "path()"    # Run the script

        s.send(function_str.encode())
        s.close()

        # Wait for the robot to reach the last path position
        if wait:
            target = path_points[len(path_points) - 1].target
            move_done = False
            while not move_done:
                if isinstance(target, JointPosition):
                    move_done = target == self.get_joint_position()
                else:
                    move_done = target == self.get_pose()

    def get_pose(self) -> Pose:
        '''Get the robots correct pose as an instance of the Pose class.'''
        conf = rtde_config.ConfigFile(self._args.config)
        output_names, output_types = conf.get_recipe('out')

        con = rtde.RTDE(self._args.host, self._args.port)
        con.connect()

        con.get_controller_version()
        con.send_output_setup(output_names, output_types, frequency=self._args.frequency)
        con.send_start()

        if self._args.buffered:
            state = con.receive_buffered(self._args.binary)
        else:
            state = con.receive(self._args.binary)

        if state is not None:
            x, y, z, rx, ry, rz = state.actual_TCP_pose
        else:
            print("[urpy][ERROR]: Failed to get pose!")

        con.send_pause()
        con.disconnect()

        pose = Pose(x, y, z, rx, ry, rz)
        pose.to_mm()

        return pose
    
    def get_joint_position(self) -> JointPosition:
        '''Get the robots correct joint positions as an instance of the JointPosition class.'''
        conf = rtde_config.ConfigFile(self._args.config)
        output_names, output_types = conf.get_recipe('out')

        con = rtde.RTDE(self._args.host, self._args.port)
        con.connect()

        con.get_controller_version()
        con.send_output_setup(output_names, output_types, frequency=self._args.frequency)
        con.send_start()

        if self._args.buffered:
            state = con.receive_buffered(self._args.binary)
        else:
            state = con.receive(self._args.binary)

        if state is not None:
            base, shoulder, elbow, wrist1, wrist2, wrist3 = state.actual_q
        else:
            print("[urpy][ERROR]: Failed to get joint position!")

        con.send_pause()
        con.disconnect()

        joint_position = JointPosition(base, shoulder, elbow, wrist1, wrist2, wrist3)

        return joint_position


class Pose:
    '''A wrapper around an arm pose.'''
    def __init__(self, x: float = 0.0, y: float = 0.0, z: float = 0.0, rx: float = 0.0, ry: float = 0.0, rz: float = 0.0) -> None:
        self.x: float = x
        self.y: float = y
        self.z: float = z
        self.rx: float = round(rx, 2)
        self.ry: float = round(ry, 2)
        self.rz: float = round(rz, 2)

    def _get_in_m(self):
        '''Returns pose in m. The robot needs meter space cordinates.'''
        x = self.x / 1000.0
        y = self.y / 1000.0
        z = self.z / 1000.0
        return x, y, z

    def to_m(self) -> None:
        '''Converts from mm to m.'''
        self.x = self.x / 1000.0
        self.y = self.y / 1000.0
        self.z = self.z / 1000.0
    
    def to_mm(self) -> None:
        '''Converts from m to mm.'''
        self.x = round(self.x * 1000.0, 1)
        self.y = round(self.y * 1000.0, 1)
        self.z = round(self.z * 1000.0, 1)

    def to_declaration(self) -> str:
        '''Returns the declaration of the pose.'''
        return "Pose(x=" + str(self.x) + ", y=" + str(self.y) + ", z=" + str(self.z) + ", rx=" + str(self.rx) + ", ry=" + str(self.ry) + ", rz=" + str(self.rz) + ")"

    def _get_undefined_move_command(self, a, v) -> str:
        '''Helper function for sending a pose to the robot.'''
        x, y, z = self._get_in_m()
        return "p[" + str(x) + ", " + str(y) + ", " + str(z) + ", " + str(self.rx) + ", " + str(self.ry) + ", " + str(self.rz) + "],a=" + str(a) + ", v=" +str(v)

    def movej(self, a, v) -> str:
        '''Returns a function string for the pose.'''
        return "movej(" + self._get_undefined_move_command(a, v) + ")\n"

    def movel(self, a, v) -> str:
        '''Returns a function string for the pose.'''
        return "movel(" + self._get_undefined_move_command(a, v) + ")\n"
    
    def lerp(self, other: Pose, percent: float) -> Pose:
        '''Returns a linear interpolated pose.'''
        x = round(lerp(self.x, other.x, percent), 1)
        y = round(lerp(self.y, other.y, percent), 1)
        z = round(lerp(self.z, other.z, percent), 1)
        rx = round(lerp(self.rx, other.rx, percent), 2)
        ry = round(lerp(self.ry, other.ry, percent), 2)
        rz = round(lerp(self.rz, other.rz, percent), 2)
        return Pose(x, y, z, rx, ry, rz)

    def copy(self) -> Pose:
        '''Returns a copy of the pose.'''
        return Pose(x=self.x, y=self.y, z=self.z, rx=self.rx, ry=self.ry, rz=self.rz)

    def _kinda_equal(self, other: Pose, threshold: Float):
        '''Returns true if the arm is within a threshold of other.'''
        if isinstance(other, Pose):
            x = (self.x + threshold >= other.x) and (self.x - threshold <= other.x)
            y = (self.y + threshold >= other.y) and (self.y - threshold <= other.y)
            z = (self.z + threshold >= other.z) and (self.z - threshold <= other.z)
            return x and y and z
        return False

    def __eq__(self, other: Pose):
        if isinstance(other, Pose):
            # return (self.x == other.x) and (self.y == other.y) and (self.z == other.z) and (self.rx == other.rx) and (self.ry == other.ry) and (self.rz == other.rz)
            return (self.x == other.x) and (self.y == other.y) and (self.z == other.z)
        return False
    
    def __str__(self):
        return "X: " + str(self.x) + ", Y: " + str(self.y) + ", Z: " + str(self.z) + ", rX: " + str(self.rx) + ", rY: " + str(self.ry) + ", rZ: " + str(self.rz)


class JointPosition:
    '''A wrapper around a joint position. Defined in radians!'''
    def __init__(self, base: float, shoulder: float, elbow: float, wrist1: float, wrist2: float, wrist3: float) -> None:
        self.base: float = round(base, 3)
        self.shoulder: float = round(shoulder, 3)
        self.elbow: float = round(elbow, 3)
        self.wrist1: float = round(wrist1, 3)
        self.wrist2: float = round(wrist2, 3)
        self.wrist3: float = round(wrist3, 3)

    def to_declaration(self) -> str:
        '''Returns the declaration of the joint position.'''
        return "JointPosition(base=" + str(self.base) + ", shoulder=" + str(self.shoulder) + ", elbow=" + str(self.elbow) + ", wrist1=" + str(self.wrist1) + ", wrist2=" + str(self.wrist2) + ", wrist3=" + str(self.wrist3) + ")"

    def _get_undefined_move_command(self, a, v) -> str:
        '''Helper function for sending a pose to the robot.'''
        return "[" + str(self.base) + ", " + str(self.shoulder) + ", " + str(self.elbow) + ", " + str(self.wrist1) + ", " + str(self.wrist2) + ", " + str(self.wrist3) + "],a=" + str(a) + ", v=" +str(v)

    def movej(self, a, v) -> str:
        '''Returns a function string for the joint position.'''
        return "movej(" + self._get_undefined_move_command(a, v) + ")\n"

    def lerp(self, other: JointPosition, percent: float) -> Pose:
        '''Returns a linear interpolated pose.'''
        base = round(lerp(self.base, other.base, percent), 3)
        shoulder = round(lerp(self.shoulder, other.shoulder, percent), 3)
        elbow = round(lerp(self.elbow, other.elbow, percent), 3)
        wrist1 = round(lerp(self.wrist1, other.wrist1, percent), 3)
        wrist2 = round(lerp(self.wrist2, other.wrist2, percent), 3)
        wrist3 = round(lerp(self.wrist3, other.wrist3, percent), 3)
        return JointPosition(base, shoulder, elbow, wrist1, wrist2, wrist3)

    def copy(self) -> JointPosition:
        '''Returns a copy of the joint position.'''
        return JointPosition(self.base, self.shoulder, self.elbow, self.wrist1, self.wrist2, self.wrist3)

    def _kinda_equal(self, other: JointPosition, threshold: Float):
        '''Returns true if the arm is within a threshold of other.'''
        if isinstance(other, JointPosition):
            base = (self.base + threshold >= other.base) and (self.base - threshold <= other.base)
            shoulder = (self.shoulder + threshold >= other.shoulder) and (self.shoulder - threshold <= other.shoulder)
            elbow = (self.elbow + threshold >= other.elbow) and (self.elbow - threshold <= other.elbow)
            wrist1 = (self.wrist1 + threshold >= other.wrist1) and (self.wrist1 - threshold <= other.wrist1)
            wrist2 = (self.wrist2 + threshold >= other.wrist2) and (self.wrist2 - threshold <= other.wrist2)
            wrist3 = (self.wrist3 + threshold >= other.wrist3) and (self.wrist3 - threshold <= other.wrist3)
            return base and shoulder and elbow and wrist1 and wrist2 and wrist3
        return False

    def __eq__(self, other):
        if (isinstance(other, JointPosition)):
            return (self.base == other.base) and (self.shoulder == other.shoulder) and (self.elbow == other.elbow) and (self.wrist1 == other.wrist1) and (self.wrist2 == other.wrist2) and (self.wrist3 == other.wrist3)
        return False
    
    def __str__(self):
        return "Base: " + str(self.base) + ", Shoulder: " + str(self.shoulder) + ", Elbow: " + str(self.elbow) + ", Wrist1: " + str(self.wrist1) + ", Wrist2: " + str(self.wrist2) + ", Wrist3: " + str(self.wrist3)


class PathPoint:
    def __init__(self, target: Union[Pose, JointPosition], movement_type: MovementType = MovementType.QUICKEST) -> None:
        self.target: Union[Pose, JointPosition] = target
        self.movement_type: MovementType = movement_type


def lerp(a: float, b: float, percent: float) -> float:
    '''Takes two values and a percent between 0 and 1. Returns a liner interpolated value.'''
    return a + ((b - a) * percent)



import struct


class ControlHeader(object):
    __slots__ = ['command', 'size',]
    
    @staticmethod
    def unpack(buf):
        rmd = ControlHeader()
        (rmd.size, rmd.command) = struct.unpack_from('>HB', buf)
        return rmd


class ControlVersion(object):
    __slots__ = ['major', 'minor', 'bugfix', 'build']
    
    @staticmethod
    def unpack(buf):
        rmd = ControlVersion()
        (rmd.major, rmd.minor, rmd.bugfix, rmd.build) = struct.unpack_from('>IIII', buf)
        return rmd


class ReturnValue(object):
    __slots__ = ['success']
    
    @staticmethod
    def unpack(buf):
        rmd = ReturnValue()
        rmd.success = bool(struct.unpack_from('>B', buf)[0])
        return rmd

class MessageV1(object):
    @staticmethod
    def unpack(buf):
        rmd = Message() # use V2 message object
        offset = 0
        rmd.level = struct.unpack_from(">B", buf, offset)[0]
        offset = offset + 1
        rmd.message = str(buf[offset:])
        rmd.source = ""

        return rmd


class Message(object):
    __slots__ = ['level', 'message', 'source']
    EXCEPTION_MESSAGE = 0
    ERROR_MESSAGE = 1
    WARNING_MESSAGE = 2
    INFO_MESSAGE = 3
    
    @staticmethod
    def unpack(buf):
        rmd = Message()
        offset = 0
        msg_length = struct.unpack_from(">B", buf, offset)[0]
        offset = offset + 1
        rmd.message = str(buf[offset:offset+msg_length])
        offset = offset + msg_length

        src_length = struct.unpack_from(">B", buf, offset)[0]
        offset = offset + 1
        rmd.source = str(buf[offset:offset+src_length])
        offset = offset + src_length
        rmd.level = struct.unpack_from(">B", buf, offset)[0]

        return rmd


def get_item_size(data_type):
    if data_type.startswith('VECTOR6'):
        return 6
    elif data_type.startswith('VECTOR3'):
        return 3
    return 1

def unpack_field(data, offset, data_type):
    size = get_item_size(data_type)
    if(data_type == 'VECTOR6D' or
       data_type == 'VECTOR3D'):
        return [float(data[offset+i]) for i in range(size)]
    elif(data_type == 'VECTOR6UINT32'):
        return [int(data[offset+i]) for i in range(size)]
    elif(data_type == 'DOUBLE'):
        return float(data[offset])
    elif(data_type == 'UINT32' or
         data_type == 'UINT64'):
        return int(data[offset])
    elif(data_type == 'VECTOR6INT32'):
        return [int(data[offset+i]) for i in range(size)]
    elif(data_type == 'INT32' or
         data_type == 'UINT8'):
        return int(data[offset])
    elif(data_type == 'BOOL'):
        return bool(data[offset])
    raise ValueError('unpack_field: unknown data type: ' + data_type)


class DataObject(object):
    recipe_id = None
    def pack(self, names, types):
        if len(names) != len(types):
            raise ValueError('List sizes are not identical.')
        l = []
        if(self.recipe_id is not None):
            l.append(self.recipe_id)
        for i in range(len(names)):
            if self.__dict__[names[i]] is None:
                raise ValueError('Uninitialized parameter: ' + names[i])
            if types[i].startswith('VECTOR'):
                l.extend(self.__dict__[names[i]])
            else:
                l.append(self.__dict__[names[i]])
        return l
    
    @staticmethod
    def unpack(data, names, types):
        if len(names) != len(types):
            raise ValueError('List sizes are not identical.')
        obj = DataObject()
        offset = 0
        obj.recipe_id = data[0]
        for i in range(len(names)):
            obj.__dict__[names[i]] = unpack_field(data[1:], offset, types[i])
            offset += get_item_size(types[i])
        return obj

    @staticmethod
    def create_empty(names, recipe_id):
        obj = DataObject()
        for i in range(len(names)):
            obj.__dict__[names[i]] = None
        obj.recipe_id = recipe_id
        return obj


class DataConfig(object):
    __slots__ = ['id', 'names', 'types', 'fmt']
    @staticmethod
    def unpack_recipe(buf):
        rmd = DataConfig();
        rmd.id = struct.unpack_from('>B', buf)[0]
        rmd.types = buf.decode('utf-8')[1:].split(',')
        rmd.fmt = '>B'
        for i in rmd.types:
            if i=='INT32':
                rmd.fmt += 'i'
            elif i=='UINT32':
                rmd.fmt += 'I'
            elif i=='VECTOR6D':
                rmd.fmt += 'd'*6
            elif i=='VECTOR3D':
                rmd.fmt += 'd'*3
            elif i=='VECTOR6INT32':
                rmd.fmt += 'i'*6
            elif i=='VECTOR6UINT32':
                rmd.fmt += 'I'*6
            elif i=='DOUBLE':
                rmd.fmt += 'd'
            elif i=='UINT64':
                rmd.fmt += 'Q'
            elif i=='UINT8':
                rmd.fmt += 'B'
            elif i =='BOOL':
                rmd.fmt += '?'
            elif i=='IN_USE':
                raise ValueError('An input parameter is already in use.')
            else:
                raise ValueError('Unknown data type: ' + i)
        return rmd
        
    def pack(self, state):
        l = state.pack(self.names, self.types)
        return struct.pack(self.fmt, *l)

    def unpack(self, data):
        li =  struct.unpack_from(self.fmt, data)
        return DataObject.unpack(li, self.names, self.types)
    
